@@ -24,7 +24,10 @@
 #include <sys/types.h>
 #include <errno.h>
 #include <stdio.h>
+#include <stdlib.h>
 #include <pthread.h>
+#include <unistd.h>
+#include <sys/fsuid.h>
 
 #include "Cipher.h"
 #include "Mutex.h"
@@ -264,6 +267,9 @@ std::string DirTraverse::nextInvalid()
 DirNode::DirNode(const string &sourceDir, const Ptr<Config> &_config)
 {
     pthread_mutex_init( &mutex, 0 );
+    
+    Lock _lock( mutex );
+
     rootDir = sourceDir;
     config = _config;
 
@@ -300,7 +306,7 @@ DirNode::~DirNode()
 
     if(!openFiles.empty())
     {
-	rError("Leaked FileNodes: %i", openFiles.size());
+	rError("Leaked FileNodes: %i", (int)openFiles.size());
 	CacheMap::iterator it = openFiles.begin();
 	for(; it!= openFiles.end(); ++it)
 	{
@@ -372,7 +378,7 @@ DirNode::relativeCipherPath( const char *plaintextPath )
 
     try
     {
-	if(plaintextPath[0] == '/' && plaintextPath[1] != '\0')
+	if(plaintextPath[0] == '/')
 	{
 	    // mark with '+' to indicate special decoding..
 	    return string("+") + naming->encodeName(plaintextPath+1, 
@@ -623,19 +629,34 @@ bool DirNode::recursiveRename( const char *fromP, const char *toP )
     return ok;
 }
 
-int DirNode::mkdir(const char *plaintextPath, mode_t mode)
+int DirNode::mkdir(const char *plaintextPath, mode_t mode, 
+	uid_t uid, gid_t gid)
 {
     string cyName = rootDir + naming->encodePath( plaintextPath );
     rAssert( !cyName.empty() );
 
     rLog( Info, "mkdir on %s", cyName.c_str() );
 
+    // if uid or gid are set, then that should be the directory owner
+    int olduid = -1;
+    int oldgid = -1;
+    if(uid != 0)
+	olduid = setfsuid( uid );
+    if(gid != 0)
+	oldgid = setfsgid( gid );
+
     int res = ::mkdir( cyName.c_str(), mode );
+
+    if(olduid >= 0)
+	setfsuid( olduid );
+    if(oldgid >= 0)
+	setfsgid( oldgid );
+
     if(res == -1)
     {
 	int eno = errno;
-	rWarning("mkdir error on %s (%s) mode %i: %s", plaintextPath, 
-		cyName.c_str(), mode, strerror(eno));
+	rWarning("mkdir error on %s mode %i: %s", cyName.c_str(), 
+		mode, strerror(eno));
 	res = -eno;
     } else
 	res = 0;
@@ -997,9 +1018,7 @@ FileNode *DirNode::findOrCreate( const char *plainName, const char *requestor,
 
     if(node)
     {	
-	pthread_mutex_lock( &node->mutex );
-	++node->refCnt;
-	pthread_mutex_unlock( &node->mutex );
+	node->incRef();
 	*created = false;
     } else
     {
@@ -1042,9 +1061,7 @@ DirNode::lookupNode( const char *plainName, const char * requestor )
 	if( recentNodes[ nextCacheLoc ] )
 	    FileNodeDestructor( recentNodes[ nextCacheLoc ] );
 
-	pthread_mutex_lock( &node->mutex );
-	++node->refCnt;
-	pthread_mutex_unlock( &node->mutex );
+	node->incRef();
 
 	recentNodes[ nextCacheLoc ] = node;
 	if(++nextCacheLoc >= RecentNodeCacheSize)
@@ -1076,9 +1093,7 @@ DirNode::openNode( const char *plainName, const char * requestor, int flags,
 
     if( (*result = node->open( flags )) >= 0 )
     {
-	pthread_mutex_lock( &node->mutex );
-	++node->retainCount;
-	pthread_mutex_unlock( &node->mutex );
+	node->incRetain();
     
 	if(!inMap)
 	{
@@ -1108,7 +1123,6 @@ int DirNode::unlink( const char *plaintextName )
     CacheMap::const_iterator it = openFiles.find( HASHKEY(plaintextName) );
     if(it != openFiles.end())
     {
-#ifndef HAVE_FUSE_HIDE
 	// If we aren't using a version of FUSE with file-hide support, then we
 	// can't allow the unlink of an open file!
 	/*
@@ -1124,15 +1138,6 @@ int DirNode::unlink( const char *plaintextName )
 		it->second->cipherName(),
 		it->second->refCnt, it->second->retainCount);
 	res = -EBUSY;
-#else
-	FileNode *node = it->second;
-	rAssert(node != NULL);
-	if(!node->hidden)
-	{
-	    rError("Expected hidden file: %s", node->cipherName());
-	    res = -EBUSY;
-	}
-#endif
     }
 
     if(res == 0)
