@@ -20,6 +20,7 @@
 
 #include <inttypes.h>
 #include <dirent.h>
+#include <sys/types.h>
 
 #include <map>
 #include <list>
@@ -155,7 +156,9 @@ public:
     // traverse directory
     DirTraverse openDir( const char *plainDirName );
 
-    int mkdir( const char *plaintextPath, mode_t mode );
+    // uid and gid are used as the directory owner, only if not zero
+    int mkdir( const char *plaintextPath, mode_t mode,
+	    uid_t uid = 0, gid_t gid = 0);
 
     int rename( const char *fromPlaintext, const char *toPlaintext );
 
