@@ -23,6 +23,7 @@
 #include "config.h"
 
 #include "readpassphrase.h"
+#include "autosprintf.h"
 
 #include "FileUtils.h"
 #include "Config.h"
@@ -54,6 +55,7 @@
 using namespace rel;
 using namespace rlog;
 using namespace std;
+using namespace gnu;
 
 static const int DefaultBlockSize = 512;
 // environment variable names for values encfs stores in the environment when
@@ -72,21 +74,22 @@ struct ConfigInfo
 {
     const char *fileName;
     ConfigType type;
+    const char *environmentOverride;
     bool (*loadFunc)(const char *fileName, EncFSConfig *config,
 	    ConfigInfo *cfg);
     bool (*saveFunc)(const char *fileName, EncFSConfig *config);
     int currentSubVersion;
     int defaultSubVersion;
 } ConfigFileMapping[] = {
-    {".encfs5", Config_V5, readV5Config, writeV5Config, 
+    {".encfs5", Config_V5, "ENCFS5_CONFIG", readV5Config, writeV5Config, 
 	V5SubVersion, V5SubVersionDefault },
     // backward compatible support for config versions 3 and 4
-    {".encfs4", Config_V4, readV4Config, writeV4Config, 0, 0 },
-    {".encfs3", Config_V3, readV3Config, writeV3Config, 0, 0 },
+    {".encfs4", Config_V4, NULL, readV4Config, writeV4Config, 0, 0 },
+    {".encfs3", Config_V3, NULL, readV3Config, writeV3Config, 0, 0 },
     // forget about version 2 and 1 - from before the first public release!
-    {".encfs2", Config_Prehistoric, NULL, NULL, 0, 0 },
-    {".encfs", Config_Prehistoric, NULL, NULL, 0, 0 },
-    {NULL,Config_None, NULL, NULL, 0, 0}};
+    {".encfs2", Config_Prehistoric, NULL, NULL, NULL, 0, 0 },
+    {".encfs", Config_Prehistoric, NULL, NULL, NULL, 0, 0 },
+    {NULL,Config_None, NULL, NULL, NULL, 0, 0}};
 
 
 EncFS_Root::EncFS_Root()
@@ -137,34 +140,72 @@ const char *lastPathElement( const char *name )
     return loc ? loc + 1 : name;
 }
 
+bool userAllowMkdir( const char *path, mode_t mode )
+{
+    // TODO: can we internationalize the y/n names?  Seems strange to prompt in
+    // their own language but then have to respond 'y' or 'n'.
+    // xgroup(setup)
+    cerr << autosprintf( _("The directory \"%s\" does not exist. Should it be created? (y,n) "), path );
+    char answer[10];
+    fgets( answer, sizeof(answer), stdin );
+
+    if(toupper(answer[0]) == 'Y')
+    {
+	int result = mkdir( path, mode );
+	if(result < 0)
+	{
+	    perror( _("Unable to create directory: ") );
+	    return false;
+	} else
+	    return true;
+    } else
+    {
+	// Directory not created, by user request
+	cerr << _("Directory not created.") << "\n";
+	return false;
+    }
+}
+
+ConfigType readConfig_load( ConfigInfo *nm, const char *path, 
+	EncFSConfig *config )
+{
+    if( nm->loadFunc )
+    {
+	try
+	{
+	    if( (*nm->loadFunc)( path, config, nm ))
+		return nm->type;
+	} catch( rlog::Error & err )
+	{
+	    err.log( _RLWarningChannel );
+	}
+
+	rError( _("Found config file %s, but failed to load"), path);
+	return Config_None;
+    } else
+    {
+	// No load function - must be an unsupported type..
+	return nm->type;
+    }
+}
+
 ConfigType readConfig( const string &rootDir, EncFSConfig *config )
 {
     ConfigInfo *nm = ConfigFileMapping;
     while(nm->fileName)
     {
+	// allow environment variable to override default config path
+	if( nm->environmentOverride != NULL )
+	{
+	    char *envFile = getenv( nm->environmentOverride );
+	    if( envFile != NULL )
+		return readConfig_load( nm, envFile, config );
+	}
+	// the standard place to look is in the root directory
 	string path = rootDir + nm->fileName;
 	if( fileExists( path.c_str() ) )
-	{
-	    if( nm->loadFunc )
-	    {
-		try
-		{
-		    if( (*nm->loadFunc)( path.c_str(), config, nm ))
-			return nm->type;
-		} catch( rlog::Error & err )
-		{
-		    err.log( _RLWarningChannel );
-		}
+	    return readConfig_load( nm, path.c_str(), config);
 
-		rError( _("Found config file %s, but failed to load"),
-			path.c_str());
-		return Config_None;
-	    } else
-	    {
-		// No load function - must be an unsupported type..
-		return nm->type;
-	    }
-	}
 	++nm;
     }
 
@@ -320,6 +361,13 @@ bool saveConfig( ConfigType type, const string &rootDir,
 	if( nm->type == type && nm->saveFunc )
 	{
 	    string path = rootDir + nm->fileName;
+	    if( nm->environmentOverride != NULL )
+	    {
+		// use environment file if specified..
+		const char *envFile = getenv( nm->environmentOverride );
+		if( envFile != NULL )
+		    path.assign( envFile );
+	    }
 
 	    try
 	    {
@@ -734,7 +782,8 @@ bool selectExternalChainedIV()
 
 RootPtr createV5Config( const std::string &rootDir, 
 	bool enableIdleTracking, bool forceDecode,
-	const std::string &passwordProgram )
+	const std::string &passwordProgram,
+	bool useStdin )
 {
     RootPtr rootInfo;
 
@@ -835,7 +884,7 @@ RootPtr createV5Config( const std::string &rootDir,
 	}
 	selectBlockMAC(&blockMACBytes, &blockMACRandBytes);
     }
-   
+
     Ptr<Cipher> cipher = Cipher::New( alg.name, keySize );
     if(cipher.isNull())
     {
@@ -867,6 +916,21 @@ RootPtr createV5Config( const std::string &rootDir,
     cout << _("Configuration finished.  The filesystem to be created has\n"
 	"the following properties:") << endl;
     showFSInfo( config );
+    
+    if( config.externalIVChaining )
+    {
+	cout << 
+	    _("-------------------------- WARNING --------------------------\n")
+	    <<
+	    _("The external initialization-vector chaining option has been\n"
+	      "enabled.  This option disables the use of hard links on the\n"
+	      "filesystem. Without hard links, some programs may not work.\n"
+	      "The programs 'mutt' and 'procmail' are known to fail.  For\n"
+	      "more information, please see the encfs mailing list.\n"
+	      "If you would like to choose another configuration setting,\n"
+	      "please press CTRL-C now to abort and start over.") << endl;
+	cout << endl;
+    }
 
     // xgroup(setup)
     cout << _("Now you will need to enter a password for your filesystem.\n"
@@ -881,7 +945,10 @@ RootPtr createV5Config( const std::string &rootDir,
 
     // get user key and use it to encode volume key
     CipherKey userKey;
-    if(passwordProgram.empty())
+    rDebug( "useStdin: %i", useStdin );
+    if(useStdin)
+	userKey = getUserKey( cipher, useStdin );
+    else if(passwordProgram.empty())
 	userKey = getNewUserKey( cipher );
     else
 	userKey = getUserKey( passwordProgram, cipher, rootDir );
@@ -1039,14 +1106,23 @@ void showFSInfo( const EncFSConfig &config )
     cout << "\n";
 }
     
-CipherKey getUserKey( const Ptr<Cipher> &cipher )
+CipherKey getUserKey( const Ptr<Cipher> &cipher, bool useStdin )
 {
     const int MaxPassBuf = 64;
     char passBuf[MaxPassBuf];
+    char *res;
 
-    // xgroup(common)
-    char *res = readpassphrase( _("EncFS Password: "),
-	    passBuf, sizeof(passBuf)-1, RPP_ECHO_OFF );
+    if( useStdin )
+    {
+	res = fgets( passBuf, sizeof(passBuf) - 1, stdin );
+	// Kill the trailing newline.
+	passBuf[ strlen(passBuf)-1 ] = '\0';
+    } else
+    {
+	// xgroup(common)
+	res = readpassphrase( _("EncFS Password: "),
+		passBuf, sizeof(passBuf)-1, RPP_ECHO_OFF );
+    }
 
     CipherKey userKey;
     if(!res)
@@ -1194,14 +1270,12 @@ CipherKey getNewUserKey( const Ptr<Cipher> &cipher )
     return userKey;
 }
 
-RootPtr initFS(const std::string &rootDir, bool enableIdleTracking,
-	bool createIfNotFound, bool checkKey, bool forceDecode,
-	const std::string &passwordProgram )
+RootPtr initFS( const EncFS_Opts *opts )
 {
     RootPtr rootInfo;
     EncFSConfig config;
 
-    if(readConfig( rootDir, &config ) != Config_None)
+    if(readConfig( opts->rootDir, &config ) != Config_None)
     {
 	// first, instanciate the cipher.
 	Ptr<Cipher> cipher = Cipher::New( config.cipherIface, config.keySize );
@@ -1219,21 +1293,23 @@ RootPtr initFS(const std::string &rootDir, bool enableIdleTracking,
 
 	// get user key
 	CipherKey userKey;
-	if(passwordProgram.empty())
-	    userKey = getUserKey( cipher );
+	rDebug( "useStdin: %i", opts->useStdin );
+	if(opts->passwordProgram.empty())
+	    userKey = getUserKey( cipher, opts->useStdin );
 	else
-	    userKey = getUserKey( passwordProgram, cipher, rootDir );
+	    userKey = getUserKey( opts->passwordProgram, 
+		    cipher, opts->rootDir );
 
 	if(userKey.isNull())
 	    return rootInfo;
 
 	// no funny stuff
-	rDebug("configuration key size = %i", config.keyData.length());
+	rDebug("configuration key size = %i", (int)config.keyData.length());
 	rDebug("cipher key size = %i", cipher->encodedKeySize());
 	rAssert( (int)config.keyData.length() == cipher->encodedKeySize() ); 
 	// decode volume key..
 	CipherKey volumeKey = cipher->readKey( 
-		(unsigned char*)config.keyData.data(), userKey, checkKey);
+		(unsigned char*)config.keyData.data(), userKey, opts->checkKey);
 	userKey.reset();
 
 	if(volumeKey.isNull())
@@ -1266,24 +1342,25 @@ RootPtr initFS(const std::string &rootDir, bool enableIdleTracking,
 	dirNodeConfig->nameCoding = nameCoder;
 	dirNodeConfig->fsSubVersion = config.subVersion;
 	dirNodeConfig->blockSize = config.blockSize;
-	dirNodeConfig->inactivityTimer = enableIdleTracking;
+	dirNodeConfig->inactivityTimer = opts->idleTracking;
 	dirNodeConfig->blockMACBytes = config.blockMACBytes;
 	dirNodeConfig->blockMACRandBytes = config.blockMACRandBytes;
 	dirNodeConfig->uniqueIV = config.uniqueIV;
 	dirNodeConfig->externalIVChaining = config.externalIVChaining;
-	dirNodeConfig->forceDecode = forceDecode;
+	dirNodeConfig->forceDecode = opts->forceDecode;
 
 	rootInfo = RootPtr( new EncFS_Root );
 	rootInfo->cipher = cipher;
 	rootInfo->volumeKey = volumeKey;
-	rootInfo->root = Ptr<DirNode>( new DirNode( rootDir, dirNodeConfig ));
+	rootInfo->root = Ptr<DirNode>( 
+		new DirNode( opts->rootDir, dirNodeConfig ));
     } else
     {
-	if(createIfNotFound)
+	if(opts->createIfNotFound)
 	{
 	    // creating a new encrypted filesystem
-	    rootInfo = createV5Config( rootDir, enableIdleTracking,
-		    forceDecode, passwordProgram );
+	    rootInfo = createV5Config( opts->rootDir, opts->idleTracking,
+		    opts->forceDecode, opts->passwordProgram, opts->useStdin );
 	}
     }
 	
