@@ -26,14 +26,18 @@ public:
     NullNameIO( );
     virtual ~NullNameIO();
 
+    virtual rel::Interface interface() const;
+
     virtual int maxEncodedNameLen( int plaintextNameLen ) const;
     virtual int maxDecodedNameLen( int encodedNameLen ) const;
 
     virtual int encodeName( const char *plaintextName, int length,
-	                     char *encodedName ) const;
+	                    uint64_t *iv, char *encodedName ) const;
     virtual int decodeName( const char *encodedName, int length,
-	                     char *plaintextName ) const;
+	                    uint64_t *iv, char *plaintextName ) const;
 
+    // hack to help with static builds
+    static bool Enabled();
 private:
 };
 
