@@ -18,6 +18,7 @@
 
 #include "encfs.h"
 #include "config.h"
+#include "autosprintf.h"
 
 #include <iostream>
 #include <string>
@@ -27,6 +28,7 @@
 #include <stdio.h>
 #include <unistd.h>
 #include <sys/time.h>
+#include <errno.h>
 
 #include <getopt.h>
 
@@ -46,75 +48,72 @@
 #define NO_DES
 #include <openssl/ssl.h>
 #include <openssl/rand.h>
+#ifndef OPENSSL_NO_ENGINE
+#include <openssl/engine.h>
+#endif
 #endif
 
 #include <locale.h>
 
 #include "i18n.h"
 
-#ifndef MIN
-#define MIN(a,b) ((a)<(b)?(a):(b))
+#ifndef MAX
+inline static int MAX(int a, int b)
+{
+    return (a > b) ? a : b;
+}
 #endif
 
 using namespace std;
 using namespace rlog;
 using namespace rel;
+using namespace gnu;
 
 // Maximum number of arguments that we're going to pass on to fuse.  Doesn't
 // affect how many arguments we can handle, just how many we can pass on..
-const int MaxFuseArgs = 16;
+const int MaxFuseArgs = 32;
 struct EncFS_Args
 {
-    string rootDir; // where raw files are found
     string mountPoint; // where to make filesystem visible
-    string passwordProgram; // program to provide password
     bool isDaemon; // true == spawn in background, log to syslog
     bool isThreaded; // true == threaded
     bool isVerbose; // false == only enable warning/error messages
     int idleTimeout; // 0 == idle time in minutes to trigger unmount
-    bool keyCheck; // true == check key checksum
-    bool forceDecode; // don't abort on MAC checksum failures for data blocks
     const char *fuseArgv[MaxFuseArgs];
     int fuseArgc;
+
+    EncFS_Opts opts;
+
+    // for debugging
+    // In case someone sends me a log dump, I want to know how what options are
+    // in effect.  Not internationalized, since it is something that is mostly
+    // useful for me!
+    string toString()
+    {
+	ostringstream ss;
+	ss << (isDaemon ? "(daemon) " : "(fg) ");
+	ss << (isThreaded ? "(threaded) " : "(UP) ");
+	if(idleTimeout > 0)
+	    ss << "(timeout " << idleTimeout << ") ";
+	if(opts.checkKey) ss << "(keyCheck) ";
+	if(opts.forceDecode) ss << "(forceDecode) ";
+	if(opts.ownerCreate) ss << "(ownerCreate) ";
+	if(opts.useStdin) ss << "(useStdin) ";
+	for(int i=0; i<fuseArgc; ++i)
+	    ss << fuseArgv[i] << ' ';
+
+	return ss.str();
+    }
 };
 
 /* Yeah, we can only deal with one filesystem at a time... */
+// This is a limitation of encfs's fuse connection layer, not encfs.
 DirNode *FSRoot = NULL;
+bool gPublicFilesystem = false;
+EncFS_Args *FSArgs = NULL;
 
 static int oldStderr = STDERR_FILENO;
 
-/*  Ask the user for permission to create directory.  If they say ok, then do
- *  it.
-*/
-static 
-bool userAllowCreate( const char *fileName )
-{
-    // TODO: can we internationalize the y/n name??
-    // xgroup(setup)
-    cerr << autosprintf( _("The directory \"%s\" does not exist. Should it be created? (y,n) "), fileName );
-    char answer[10];
-    fgets( answer, sizeof(answer), stdin );
-
-    if(toupper(answer[0]) == 'Y')
-    {
-	int result = mkdir( fileName, 0700 );
-	if(result < 0)
-	{
-	    // mkdir failure..
-	    perror( _("Unable to create directory: ") );
-	    return false;
-	} else
-	{
-	    return true;
-	}
-    } else
-    {
-	// Directory not created, by user request -- aborting...
-	cerr << _("Directory not created.") << "\n";
-	return false;
-    }
-}
-
 static
 void usage(const char *name)
 {
@@ -137,6 +136,8 @@ void usage(const char *name)
 	"  --anykey\t\t"        "Do not verify correct key is being used\n"
 	"  --forcedecode\t\t"   "decode data even if an error is detected\n"
 	                  "\t\t\t(for filesystems using MAC block headers)\n")
+	<< _("  --public\t\t"   "act as a typical multi-user filesystem\n"
+	                  "\t\t\t(encfs must be run as root)\n")
 
 	// xgroup(usage)
 	<< _("  --extpass=program\tUse external program for password prompt\n"
@@ -161,37 +162,40 @@ void FuseUsage()
     fuse_main( argc, const_cast<char**>(argv), (fuse_operations*)NULL);
 }
 
-// in case someone sends me a log dump, I want to know how what options are in
-// effect.  I'd rather not internationalize this, since it is something that is
-// mostly useful for me!
-string showFuseArgs( EncFS_Args &args )
-{
-    ostringstream ss;
-    ss << (args.isDaemon ? "(daemon) " : "(fg) ");
-    ss << (args.isThreaded ? "(threaded) " : "(UP) ");
-    if(args.idleTimeout > 0)
-	ss << "(timeout " << args.idleTimeout << ") ";
-    for(int i=0; i<args.fuseArgc; ++i)
-	ss << args.fuseArgv[i] << ' ';
-
-    return ss.str();
-}
-
 #define PUSHARG(ARG) \
 rAssert(out->fuseArgc < MaxFuseArgs); \
 out->fuseArgv[out->fuseArgc++] = ARG
 
+static
+string slashTerminate( const string &src )
+{
+    string result = src;
+    if( result[ result.length()-1 ] != '/' )
+	result.append( "/" );
+    return result;
+}
+
 static 
 bool processArgs(int argc, char *argv[], EncFS_Args *out)
 {
     // set defaults
     out->isDaemon = true;
+#if defined(HAVE_ATOMIC_GCC) || defined(HAVE_ATOMIC_GCC_PRIVATE)
     out->isThreaded = true;
+#else
+    // no atomic update methods -- safest to stick to single-threaded mode..
+    out->isThreaded = false;
+#endif
     out->isVerbose = false;
     out->idleTimeout = 0;
-    out->keyCheck = true;
     out->fuseArgc = 0;
-    out->forceDecode = false;
+    out->opts.idleTracking = false;
+    out->opts.checkKey = true;
+    out->opts.forceDecode = false;
+    out->opts.ownerCreate = false;
+    out->opts.useStdin = false;
+    
+    bool useDefaultFlags = true;
 
     // pass executable name through
     out->fuseArgv[0] = lastPathElement(argv[0]);
@@ -204,14 +208,19 @@ bool processArgs(int argc, char *argv[], EncFS_Args *out)
    
     // TODO: can flags be internationalized?
     static struct option long_options[] = {
-	{"verbose", 0, 0, 'v'}, // verbose mode
 	{"fuse-debug", 0, 0, 'd'}, // Fuse debug mode
+	{"forcedecode", 0, 0, 'D'}, // force decode
+	// {"foreground", 0, 0, 'f'}, // foreground mode (no daemon)
+	{"fuse-help", 0, 0, 'H'}, // fuse_mount usage
 	{"idle", 1, 0, 'i'}, // idle timeout
 	{"anykey", 0, 0, 'k'}, // skip key checks
-	{"forcedecode", 0, 0, 'D'}, // force decode
+	{"no-default-flags", 0, 0, 'N'}, // don't use default fuse flags
+	{"public", 0, 0, 'P'}, // public mode
 	{"extpass", 1, 0, 'p'}, // external password program
+	// {"single-thread", 0, 0, 's'}, // single-threaded mode
+	{"stdinpass", 1, 0, 'S'}, // read password from stdin
+	{"verbose", 0, 0, 'v'}, // verbose mode
 	{"version", 0, 0, 'V'}, //version
-	{"fuse-help", 0, 0, 'H'}, // fuse_mount usage
 	{0,0,0,0}
     };
 
@@ -224,7 +233,8 @@ bool processArgs(int argc, char *argv[], EncFS_Args *out)
 	// 'v' : verbose mode (same as --verbose)
 	// 'd' : fuse debug mode (same as --fusedebug)
 	// 'i' : idle-timeout, takes argument
-	int res = getopt_long( argc, argv, "Hsfvdi:",
+	// 'S' : password from stdin
+	int res = getopt_long( argc, argv, "HsSfvdi:",
 		long_options, &option_index);
 
 	if(res == -1)
@@ -234,7 +244,9 @@ bool processArgs(int argc, char *argv[], EncFS_Args *out)
 	{
 	case 's':
 	    out->isThreaded = false;
-	    PUSHARG("-s");
+	    break;
+	case 'S':
+	    out->opts.useStdin = true;
 	    break;
 	case 'f':
 	    out->isDaemon = false;
@@ -248,16 +260,38 @@ bool processArgs(int argc, char *argv[], EncFS_Args *out)
 	    PUSHARG("-d");
 	    break;
 	case 'i':
+#if FUSE_MAJOR_VERSION == 2 && FUSE_MINOR_VERSION == 2
+	    cerr << "Sorry, idle monitoring not enabled. "
+		"FUSE 2.3 or newer required.\n" 
+		<< endl;
+#else
 	    out->idleTimeout = strtol( optarg, (char**)NULL, 10);
+	    out->opts.idleTracking = true;
+#endif
 	    break;
 	case 'k':
-	    out->keyCheck = false;
+	    out->opts.checkKey = false;
 	    break;
 	case 'D':
-	    out->forceDecode = true;
+	    out->opts.forceDecode = true;
+	    break;
+	case 'N':
+	    useDefaultFlags = false;
 	    break;
 	case 'p':
-	    out->passwordProgram.assign( optarg );
+	    out->opts.passwordProgram.assign( optarg );
+	    break;
+	case 'P':
+	    if(geteuid() != 0)
+		rWarning(_("option '--public' ignored for non-root user"));
+	    else
+	    {
+		out->opts.ownerCreate = true;
+		// add 'allow_other' option
+		// add 'default_permissions' option (default)
+		PUSHARG("-o");
+		PUSHARG("allow_other");
+	    }
 	    break;
 	case 'V':
 	    // xgroup(usage)
@@ -279,34 +313,24 @@ bool processArgs(int argc, char *argv[], EncFS_Args *out)
 	    break;
 	}
     }
+
+    if(!out->isThreaded)
+	PUSHARG("-s");
+
+    if(useDefaultFlags)
+    {
+	PUSHARG("-o");
+	PUSHARG("use_ino");
+	PUSHARG("-o");
+	PUSHARG("default_permissions");
+    }
 	    
     // we should have at least 2 arguments left over - the source directory and
     // the mount point.
     if(optind+2 <= argc)
     {
-	if(!isDirectory( argv[optind] ) && !userAllowCreate( argv[optind] ))
-	{
-	    // root directory doesn't exist, and the user refused to allow us
-	    // to create a non-existant one.
-	    rWarning(_("Unable to locate root directory, aborting."));
-	    return false;
-	} else
-	{
-	    out->rootDir = argv[optind++];
-	    if( out->rootDir[ out->rootDir.length()-1 ] != '/' )
-		out->rootDir.append( "/" );
-	}
-		
-	if(!isDirectory( argv[optind] ) && !userAllowCreate( argv[optind] ))
-	{
-	    // mount point doesn't exist, and the user refused to allow it to
-	    // be created..
-	    rWarning(_("Unable to locate mount point, aborting."));
-	    return false;
-	} else
-	{
-	    out->mountPoint = argv[optind++];
-	}
+	out->opts.rootDir = slashTerminate( argv[optind++] );
+	out->mountPoint = argv[optind++];
     } else
     {
 	// no mount point specified
@@ -330,7 +354,7 @@ bool processArgs(int argc, char *argv[], EncFS_Args *out)
     // sanity check
     if(out->isDaemon && 
 	    (!isAbsolutePath( out->mountPoint.c_str() ) ||
-	    !isAbsolutePath( out->rootDir.c_str() ) ) 
+	    !isAbsolutePath( out->opts.rootDir.c_str() ) ) 
       )
     {
 	cerr << 
@@ -341,6 +365,36 @@ bool processArgs(int argc, char *argv[], EncFS_Args *out)
 	return false;
     }
 
+    // the raw directory may not be a subdirectory of the mount point.
+    {
+	string testMountPoint = slashTerminate( out->mountPoint );
+	string testRootDir = 
+	    out->opts.rootDir.substr(0, testMountPoint.length());
+
+	if( testMountPoint == testRootDir )
+	{
+	    cerr << 
+		// xgroup(usage)
+		_("The raw directory may not be a subdirectory of the "
+		  "mount point.") << endl;
+	    return false;
+	}
+    }
+
+    // check that the directories exist, or that we can create them..
+    if(!isDirectory( out->opts.rootDir.c_str() ) && 
+	    !userAllowMkdir( out->opts.rootDir.c_str() ,0700))
+    {
+	rWarning(_("Unable to locate root directory, aborting."));
+	return false;
+    }
+    if(!isDirectory( out->mountPoint.c_str() ) && 
+	    !userAllowMkdir( out->mountPoint.c_str(),0700))
+    {
+	rWarning(_("Unable to locate mount point, aborting."));
+	return false;
+    }
+
     // fill in mount path for fuse
     out->fuseArgv[1] = out->mountPoint.c_str();
 
@@ -393,6 +447,17 @@ void pthreads_locking_cleanup()
 }
 #endif
 
+struct EncfsFSContext
+{
+    EncFS_Args *args;
+
+    // for idle monitor
+    bool running;
+    pthread_t monitorThread;
+    pthread_cond_t wakeupCond;
+    pthread_mutex_t wakeupMutex;
+};
+
 /*
     Idle monitoring thread.  This is only used when idle monitoring is enabled.
     It will cause the filesystem to be automatically unmounted (causing us to
@@ -400,20 +465,19 @@ void pthreads_locking_cleanup()
     checked if there are no open files, as I don't want to risk problems by
     having the filesystem unmounted from underneath open files!
 */
-static bool gRunning = true;
-pthread_cond_t gWakeupCond = PTHREAD_COND_INITIALIZER;
-pthread_mutex_t gWakeupMutex = PTHREAD_MUTEX_INITIALIZER;
 const int MinActivityCheckInterval = 10;
 
+static
 void * idleMonitor(void *_arg)
 {
-    EncFS_Args *arg = (EncFS_Args*)_arg;
+    EncfsFSContext *ctx = (EncfsFSContext*)_arg;
+    EncFS_Args *arg = ctx->args;
 
     int timeoutSeconds = 60 * arg->idleTimeout;
 
-    pthread_mutex_lock( &gWakeupMutex );
-
-    while(gRunning)
+    pthread_mutex_lock( &ctx->wakeupMutex );
+    
+    while(ctx->running)
     {
 	int nextCheck = MinActivityCheckInterval;
 
@@ -428,7 +492,7 @@ void * idleMonitor(void *_arg)
 			arg->mountPoint.c_str());
 		fuse_unmount( arg->mountPoint.c_str() );
 		// wait for main thread to wake us up
-		pthread_cond_wait( &gWakeupCond, &gWakeupMutex );
+		pthread_cond_wait( &ctx->wakeupCond, &ctx->wakeupMutex );
 		break;
 	    } else
 	    {
@@ -437,7 +501,7 @@ void * idleMonitor(void *_arg)
 	} else
 	{
 	    // don't need to check again until it could possibly timeout
-	    nextCheck = MIN(timeoutSeconds - idleSeconds,
+	    nextCheck = MAX(timeoutSeconds - idleSeconds,
 		            MinActivityCheckInterval);
 	    rDebug("checking idle time again in %i seconds", nextCheck);
 	}
@@ -447,16 +511,62 @@ void * idleMonitor(void *_arg)
 	struct timespec wakeupTime;
 	wakeupTime.tv_sec = currentTime.tv_sec + nextCheck;
 	wakeupTime.tv_nsec = currentTime.tv_usec * 1000;
-	pthread_cond_timedwait( &gWakeupCond, &gWakeupMutex, &wakeupTime );
+	pthread_cond_timedwait( &ctx->wakeupCond, 
+		&ctx->wakeupMutex, &wakeupTime );
     }
     
-    pthread_mutex_unlock( &gWakeupMutex );
+    pthread_mutex_unlock( &ctx->wakeupMutex );
 
     rDebug("Idle monitoring thread exiting");
 
     return 0;
 }
 
+void *encfs_init()
+{
+    EncfsFSContext *ctx = new EncfsFSContext;
+
+    ctx->args = FSArgs;
+
+    // if an idle timeout is specified, then setup a thread to monitor the
+    // filesystem.
+    if(FSArgs->idleTimeout > 0)
+    {
+	rDebug("starting idle monitoring thread");
+	ctx->running = true;
+	pthread_cond_init( &ctx->wakeupCond, 0 );
+	pthread_mutex_init( &ctx->wakeupMutex, 0 );
+
+	int res = pthread_create( &ctx->monitorThread, 0, idleMonitor, 
+		(void*)ctx );
+	if(res != 0)
+	{
+	    rError("error starting idle monitor thread, "
+		    "res = %i, errno = %i", res, errno);
+	}
+    }
+
+    return (void*)ctx;
+}
+ 
+void encfs_destroy( void *_ctx )
+{
+    EncfsFSContext *ctx = (EncfsFSContext*)_ctx;
+    if(ctx->args->idleTimeout > 0)
+    {
+	ctx->running = false;
+
+	// wake up the thread if it is waiting..
+	rDebug("waking up monitoring thread");
+	pthread_mutex_lock( &ctx->wakeupMutex );
+	pthread_cond_signal( &ctx->wakeupCond );
+	pthread_mutex_unlock( &ctx->wakeupMutex );
+	rDebug("joining with idle monitoring thread");
+	pthread_join( ctx->monitorThread , 0 );
+	rDebug("join done");
+    }
+}
+
 int main(int argc, char *argv[])
 {
     // initialize the logging library
@@ -476,6 +586,14 @@ int main(int argc, char *argv[])
     unsigned int randSeed = 0;
     RAND_bytes( (unsigned char*)&randSeed, sizeof(randSeed) );
     srand( randSeed );
+
+#ifndef OPENSSL_NO_ENGINE
+    // initialize hardware crypto engine support
+    ENGINE_load_builtin_engines();
+    ENGINE_register_all_ciphers();
+    ENGINE_register_all_digests();
+    ENGINE_register_all_RAND();
+#endif // NO_ENGINE
 #endif
 
     // log to stderr by default..
@@ -488,17 +606,17 @@ int main(int argc, char *argv[])
 
     // anything that comes from the user should be considered tainted until
     // we've processed it and only allowed through what we support.
-    EncFS_Args encfsArgs;
+    EncFS_Args *encfsArgs = new EncFS_Args;
     for(int i=0; i<MaxFuseArgs; ++i)
-	encfsArgs.fuseArgv[i] = NULL; // libfuse expects null args..
+	encfsArgs->fuseArgv[i] = NULL; // libfuse expects null args..
 
-    if(argc == 1 || !processArgs(argc, argv, &encfsArgs))
+    if(argc == 1 || !processArgs(argc, argv, encfsArgs))
     {
 	usage(argv[0]);
 	return EXIT_FAILURE;
     }
 
-    if(encfsArgs.isVerbose)
+    if(encfsArgs->isVerbose)
     {
 	// subscribe to more logging channels..
 	slog->subscribeTo( GetGlobalChannel("info") );
@@ -506,7 +624,7 @@ int main(int argc, char *argv[])
     }
   
 #ifdef HAVE_SSL
-    if(encfsArgs.isThreaded)
+    if(encfsArgs->isThreaded)
     {
 	// provide locking functions to OpenSSL since we'll be running with
 	// threads accessing openssl in parallel.
@@ -515,8 +633,8 @@ int main(int argc, char *argv[])
     }
 #endif
 
-    rDebug("Root directory: %s", encfsArgs.rootDir.c_str());
-    rDebug("Fuse arguments: %s", showFuseArgs( encfsArgs ).c_str());
+    rDebug("Root directory: %s", encfsArgs->opts.rootDir.c_str());
+    rDebug("Fuse arguments: %s", encfsArgs->toString().c_str());
 	
     fuse_operations encfs_oper;
     // in case this code is compiled against a newer FUSE library and new
@@ -542,6 +660,7 @@ int main(int argc, char *argv[])
     encfs_oper.read = encfs_read;
     encfs_oper.write = encfs_write;
     encfs_oper.statfs = encfs_statfs;
+    //encfs_oper.flush = encfs_flush;
     encfs_oper.release = encfs_release;
     encfs_oper.fsync = encfs_fsync;
 #ifdef HAVE_XATTR
@@ -550,18 +669,25 @@ int main(int argc, char *argv[])
     encfs_oper.listxattr = encfs_listxattr;
     encfs_oper.removexattr = encfs_removexattr;
 #endif // HAVE_XATTR
+#if FUSE_MINOR_VERSION >= 3 || FUSE_MAJOR_VERSION > 2
+    // .init and .destroy added in fuse 2.3
+    // encfs requires fuse 2.2 at a minimum..
+    encfs_oper.init = encfs_init;
+    encfs_oper.destroy = encfs_destroy;
+#endif
+
+    gPublicFilesystem = encfsArgs->opts.ownerCreate;
+    RootPtr rootInfo = initFS( &encfsArgs->opts );
+    
+    int returnCode = EXIT_FAILURE;
 
-    const bool enableIdleChecking = (encfsArgs.idleTimeout > 0);
-    const bool createFSIfNotFound = true;
-    RootPtr rootInfo = initFS( encfsArgs.rootDir, enableIdleChecking,
-	    createFSIfNotFound, encfsArgs.keyCheck, encfsArgs.forceDecode,
-	    encfsArgs.passwordProgram );
     if( !rootInfo.isNull() )
     {
 	// set the globally visible root directory node
 	FSRoot = rootInfo->root.get();
+	FSArgs = encfsArgs;
 	    
-	if(encfsArgs.isThreaded == false && encfsArgs.idleTimeout > 0)
+	if(encfsArgs->isThreaded == false && encfsArgs->idleTimeout > 0)
 	{
 	    // xgroup(usage)
 	    cerr << _("Note: requested single-threaded mode, but an idle\n"
@@ -574,7 +700,7 @@ int main(int argc, char *argv[])
 	// pass-thru calls..
 	umask( 0 );
 
-	if(encfsArgs.isDaemon)
+	if(encfsArgs->isDaemon)
 	{
 	    // switch to logging just warning and error messages via syslog
 	    logNode = new SyslogNode( "encfs" );
@@ -590,24 +716,27 @@ int main(int argc, char *argv[])
 	    oldStderr = dup( STDERR_FILENO );
 	}
 
-	// if an idle timeout is specified, then setup a thread to monitor the
-	// filesystem.
-	pthread_t monitorThread;
-	if(encfsArgs.idleTimeout > 0)
-	{
-	    rDebug("starting idle monitoring thread");
-	    pthread_create( &monitorThread, 0, idleMonitor, 
-		    (void*)&encfsArgs );
-	}
-
 	try
 	{
-	    // fuse_main returns an error code in newer versions of fuse..
-	    int res = fuse_main( encfsArgs.fuseArgc, 
-		    const_cast<char**>(encfsArgs.fuseArgv), 
-		    &encfs_oper);
+	    time_t startTime, endTime;
+	   
+	    // FIXME: workaround for fuse_main returning an error on normal
+	    // exit.  Only print information if fuse_main returned
+	    // immediately..
+	    time( &startTime );
 
-	    if(res != 0 && encfsArgs.isDaemon)
+	    // fuse_main returns an error code in newer versions of fuse..
+	    int res = fuse_main( encfsArgs->fuseArgc, 
+		    const_cast<char**>(encfsArgs->fuseArgv), 
+		    &encfs_oper);
+	    
+	    time( &endTime );
+
+	    if(res == 0)
+		returnCode = EXIT_SUCCESS;
+
+	    if(res != 0 && encfsArgs->isDaemon
+		    && (endTime - startTime <= 1) )
 	    {
 		// the users will not have seen any message from fuse, so say a
 		// few words in libfuse's memory..
@@ -617,7 +746,6 @@ int main(int argc, char *argv[])
 			" - fuse kernel module not installed (modprobe fuse)\n"
 			" - invalid options -- see usage message\n"));
 		fclose(out);
-		exit(1);
 	    }
 	} catch(std::exception &ex)
 	{
@@ -627,20 +755,6 @@ int main(int argc, char *argv[])
 	{
 	    rError(_("Internal error: Caught unexpected exception"));
 	}
-  
-	if(encfsArgs.idleTimeout > 0)
-	{
-	    gRunning = false;
-
-	    // wake up the thread if it is waiting..
-	    rDebug("waking up monitoring thread");
-	    pthread_mutex_lock( &gWakeupMutex );
-	    pthread_cond_signal( &gWakeupCond );
-	    pthread_mutex_unlock( &gWakeupMutex );
-	    rDebug("joining with idle monitoring thread");
-	    pthread_join( monitorThread , 0 );
-	    rDebug("join done");
-	}
     }
 
     // cleanup so that we can check for leaked resources..
@@ -649,10 +763,11 @@ int main(int argc, char *argv[])
 
     MemoryPool::destroyAll();
 #ifdef HAVE_SSL
-    if(encfsArgs.isThreaded)
+    if(encfsArgs->isThreaded)
 	pthreads_locking_cleanup();
 #endif
     delete logNode;
+    delete encfsArgs;
 
-    return 0;
+    return returnCode;
 }
